__author__ = 'meignier'

import copy
import logging
import numpy as np
from scipy.sparse import csgraph
from collections import namedtuple
from sidekit.bosaris.scores import Scores
from s4d.clustering.hac_utils import *


ConnectedComponentTuple = namedtuple('ConnectedComponent', ['type', 'diarization', 'scores' ])


StarGraphTuple = namedtuple('ConnectedComponent', ['type', 'diarization', 'scores',
                                                     'center', 'within_inertia'])


class ConnectedComponent:
    def __init__(self, diar, scores, threshold):
        """
        :param diarization: s4d annotation
        :param cluster_list: list of cluster_list, the ith cluster corresponds to the speaker
            of column and row i in distances_
        :param distances: matrice of distances between cluster_list (speakers)
        :param thr: the threshold will be apply to the distances_ to generate a graph
        """
        # output: the list of CC, each row contains the sub-diarization, the sub-matrix
        #  and the cluster list of the cc
        self.cc = list()
        # output: contain the star graphs (and isolated vertices)
        self.diar = copy.deepcopy(diar)
        self.scores = copy.deepcopy(scores)
        self.thr = threshold
        self.cc_nb = -1
        self.cc_list = None
        self.nb_sg = 0
        self.nb_sg0 = 0
        self.n = 0


    def _star_graph(self, lst, graph):
        """
        Verify if a star graph exists in the sub graph
        (http://en.wikipedia.org/wiki/Star_(graph_theory))
        :param lst: list of vertices
        :param graph: the graph
        :return the center and the within_inertia
        """
        within_inertia = np.inf
        center = -1
        for i in lst:
            s = np.sum(graph[i, lst])
            if within_inertia > s:
                within_inertia = s
                center = i
        return center, within_inertia

    def sub_graph(self, rename_cc=False):
        """
        find the 2 kind of sub-graphs:
         - isolated vertices and star graphs already clustered
         - other connected components put in self.cc list
        :return: generate self.cc and self.table_out
        """
        logging.debug('threshold the distance matrix')
        distances, t = scores2distance(self.scores, self.thr)

        mask = (distances>t)
        graph = distances.copy()
        graph[mask] = np.inf
        #graph = threshold(distances, threshmax=t, newval=np.inf)
        logging.debug('get connected components')
        cc_nb, cc_list = csgraph.connected_components(graph, directed=False)
        diar_out = copy.deepcopy(self.diar)
        # list of lists, each sub list contains the index of a connected component
        # print(cc_nb, cc_list)
        lst = []
        for i in range(cc_nb):
            lst.append(list())

        for j in range(len(cc_list)):
            lst[cc_list[j]].append(j)
        self.cc_list = list()
        # for each connected component
        for slst in lst:
            # start graph if the list contains only one index
            #print(slst)
            scores = Scores()
            scores.modelset = np.copy(self.scores.modelset[slst])
            scores.segset = np.copy(self.scores.segset[slst])
            scores.scoremask = np.copy(self.scores.scoremask[slst, :][:, slst])
            scores.scoremat = np.copy(self.scores.scoremat[slst, :][:, slst])

            diar_cc = self.diar.filter('cluster', 'in', scores.modelset.tolist())
            if len(slst) == 1: # isolated vertex
                cc = StarGraphTuple('star_graph_0', diar_cc, scores, self.scores.modelset[slst[0]], 0)
                self.nb_sg0 += 1
            else:
                center, within_inertia = self._star_graph(slst, graph)
                if center >= 0: # Star Graph if the list contains a center
                    l = diar_cc.unique('cluster')
                    #logging.info('nb_sg rename '+str(l)+' into '+self.scores.modelset[center])
                    diar_out.rename('cluster', l, self.scores.modelset[center])
                    diar_cc.rename('cluster', l, self.scores.modelset[center])
                    cc = StarGraphTuple('star_graph_k', diar_cc, scores, self.scores.modelset[center], within_inertia)
                    self.nb_sg += 1
                else: # connected component without star graph
                    if rename_cc:
                        diar_out.rename('cluster', diar_cc.unique('cluster'), self.scores.modelset[center])
                        #diar_cc.rename('cluster', diar_cc.unique('cluster'), self.scores.modelset[center])
                    cc = ConnectedComponentTuple('cc', diar_cc, scores)
                    self.n += 1

            self.cc_list.append(cc)
        logging.debug('-- stat CC '+str(self.thr)+' -- nb_sg0: '+str(self.nb_sg0)
                     +' nb_sg: '+str(self.nb_sg)+' cc: '+str(self.n)+ ' nb: '
                     +str(self.n + self.nb_sg0 + self.nb_sg)+'/'+str(cc_nb))
        return diar_out, self.cc_list, self.nb_sg0, self.nb_sg, self.n


def connexted_component(diar, scores, threshold):

    graphs = ConnectedComponent(diar, scores, threshold)
    return graphs.sub_graph()




import os
from ecyglpki import Problem, SimplexControls
from scipy.cluster import hierarchy as hac
import matplotlib.pyplot as plt
from scipy import stats
import logging
import time
import os
from s4d.clustering.hac_utils import *

class ILP_IV:
    def __init__(self, diar, scores, threshold=0.0):
        self.sep = '!!'
        self.dic = dict()
        self.diar = copy.deepcopy(diar)

        self.thr = threshold
        self.scores = copy.deepcopy(scores)

    def plot(self, link, cluster_list):
        #print(link)
        hac.dendrogram(link, color_threshold=self.thr, labels=cluster_list)
        plt.show()

    def perform(self, save_ilp=False):
        """
        Creates an LP problem and solves it using GLPK library 
        (with ecyglpki Cython interface)
        :param save_ilp: if True, saves the generated problem
        in CPLEX LP format (problem.ilp), and the solution of
        the problem (solution.ilp).
        :return: a Diar object with the new clusters and a 
        dictionary of clusters (old clusters as key, new ones
        as value in an array).
        """
        distances, t = scores2distance(self.scores, self.thr)
        cluster_list = sorted(self.scores.modelset.tolist())

        lp = Problem()
        lp.set_obj_dir('minimize')

        # columns
        cols = []
        for idx, cluster in enumerate(cluster_list):
            col = '{}{}{}'.format(cluster, self.sep, cluster)
            cols.append(col)
            lp.add_named_cols(col)
            lp.set_obj_coef(col, 1)
            lp.set_col_bnds(col, 0, None)

        # sum of dist > thr in the lower triangular part od the distance matrix
        mask = (np.tril(distances, -1) > t)
        threshold = np.tril(distances, -1).copy()
        threshold[mask] = 0
        s = np.sum(threshold) + 1
        l = len(cluster_list)
        for i in range(l):
            cluster_i = cluster_list[i]
            for j in range(i + 1, l):
                if distances[i, j] < t:
                    cluster_j = cluster_list[j]
                    v = distances[i, j] / s
                    col = '{}{}{}'.format(cluster_i, self.sep, cluster_j)
                    cols.append(col)
                    lp.add_named_cols(col)
                    lp.set_obj_coef(col, v)
                    lp.set_col_bnds(col, 0, None)

        # rows
        for i, cluster_i in enumerate(cluster_list):
            r_cols = {}
            row = 'S{}'.format(i)
            lp.add_named_rows(row)
            r_cols['{}{}{}'.format(cluster_i, self.sep, cluster_i)] = 1

            for j, cluster_j in enumerate(cluster_list):
                if i != j and distances[i, j] < t:
                    col = '{}{}{}'.format(cluster_i, self.sep, cluster_j)
                    if col not in cols:
                        cols.append(col)
                        lp.add_named_cols(col)
                        lp.set_col_bnds(col, 0, None)
                    r_cols[col] = 1

                    #boundaries <= 0
                    col = '{}{}{}'.format(cluster_i, self.sep, cluster_j)
                    idx = lp.add_rows(1)
                    lp.set_mat_row(idx, {'{}{}{}'.format(cluster_j, self.sep, cluster_j):-1, col:1})
                    lp.set_row_bnds(idx, None, 0)
                    
            lp.set_mat_row(row, r_cols)
            lp.set_row_bnds(row, 1, 1)

        if save_ilp:
            lp.write_lp('problem.ilp')

        # solving problem
        ctrl = SimplexControls()
        ctrl.presolve = True
        lp.simplex(ctrl)

        if save_ilp:
            lp.print_sol('solution.ilp')

        cluster_dict = dict()
        for i in range(lp.get_num_cols()):
            names = lp.get_col_name(i+1).split(self.sep)
            activity = lp.get_col_prim(i+1)
            if activity == 1 and names[0] != names[1]:
                if names[1] not in cluster_dict:
                    cluster_dict[names[1]] = []
                cluster_dict[names[1]].append(names[0])

        table = copy.deepcopy(self.diar)
        for idx in cluster_dict:
            table.rename('cluster', cluster_dict[idx], idx)
        return table, cluster_dict



    def _perform(self, filename='tmp.ilp', rm_tmp=True):
        """
        Same as perform(), using glpk solver directly
        instead of the Cython interface (slower).
        """
        table = copy.deepcopy(self.diar)
        logging.debug('ilp filename: %s', filename)
        f = open(filename, 'w')
        self._ilp_write(f)
        f.close()
        while not os.path.exists(filename):
            time.sleep(1)
        cmd = 'glpsol --lp {} -o {} &> {}'.format(filename, filename + '.out', filename + '.err')
        #print(cmd)
        if os.path.exists(filename + '.out'):
            os.remove(filename + '.out')
            while os.path.exists(filename + '.out'):
                time.sleep(1)
        os.system(cmd)
        time.sleep(1)
        while not os.path.exists(filename + '.out'):
            time.sleep(1)
        f = open(filename + '.out', 'r')
        cluster_dict = self._ilp_read(f)
        f.close()
        for idx in cluster_dict:
            table.rename('cluster', cluster_dict[idx], idx)
        if rm_tmp:
            os.remove(filename)
            os.remove(filename+ '.out')
            os.remove(filename+ '.err')
            while os.path.exists(filename) or os.path.exists(filename+ '.out') or os.path.exists(filename+ '.err'):
                time.sleep(1)
        return table, cluster_dict

    def _ilp_write(self, f):
        distances, t = scores2distance(self.scores, self.thr)

        cluster_list = self.scores.modelset.tolist()
        f.write("Minimize\n")
        f.write("problem : \n")
        for idx, cluster in enumerate(cluster_list):
            f.write(' + {}{}{}'.format(cluster, self.sep, cluster))
        # sum of dist > thr in the lower triangular part od the distance matrix

        mask = (np.tril(distances, -1)>t)
        threshold = np.tril(distances, -1).copy()
        threshold[mask] = 0
        s = np.sum(threshold) + 1
        #s = np.sum(stats.threshold(np.tril(distances, -1), threshmax=t, newval=0)) + 1
        logging.debug('ilp sum scores: '+str(s))
        l = len(cluster_list)
        for i in range(l):
            cluster_i = cluster_list[i]
            for j in range(i+1, l):
                if distances[i, j] < t:
                    cluster_j = cluster_list[j]
                    v = distances[i, j] / s
                    f.write(
                        ' +{} {}{}{} '.format(v, cluster_i, self.sep,
                                                cluster_j))
        f.write("\nSubject to\n")

        # x_i,i is a centre or not
        for i, cluster_i in enumerate(cluster_list):
            f.write('S{}: {}{}{}'.format(i, cluster_i, self.sep, cluster_i))
            for j, cluster_j in enumerate(cluster_list):
                if i != j and distances[i, j] < t:
                    f.write(" + {}{}{}".format(cluster_i, self.sep, cluster_j))
            f.write(' = 1\n')

        # center
        for i, cluster_i in enumerate(cluster_list):
            for j, cluster_j in enumerate(cluster_list):
                if i != j and distances[i, j] < t:
                    f.write("{}{}{} - {}{}{} <=0\n".format(cluster_i, self.sep, cluster_j,
                                                           cluster_j, self.sep, cluster_j))
        f.write('End')

    def _ilp_read(self, f):
        cluster_dict = dict()
        for line in f:
            line.strip()
            if line.find(self.sep) != -1:
                lst = line.split()
                if len(lst) < 3:
                    line += f.read()
                    lst = line.split()
                names = lst[1].split(self.sep)
                if lst[3] == '1' and names[0] != names[1]:
                    logging.debug('merge %s in %s', names[0], names[1])
                    if names[1] not in cluster_dict:
                        cluster_dict[names[1]] = []
                    cluster_dict[names[1]].append(names[0])
        return cluster_dict


def ilp_iv(diar, scores, threshold=0.0):
    ilp = ILP_IV(diar, scores, threshold)
    return ilp.perform()



__author__ = 'meignier'

import s4d.clustering.hac_bic
import s4d.clustering.hac_clr
import s4d.clustering.hac_iv
import s4d.clustering.hac_utils
import s4d.model_iv
from s4d.clustering.cc_iv import ConnectedComponent
from s4d.diar import Diar
from s4d.segmentation import sanity_check, bic_linear, div_gauss
from s4d.viterbi import Viterbi

__version__ = "0.0.3"



# -*- coding: utf-8 -*-
# -*- coding: utf-8 -*-
#
# This file is part of S4D.
#
# SD4 is a python package for speaker diarization based on SIDEKIT.
# S4D home page: http://www-lium.univ-lemans.fr/s4d/
# SIDEKIT home page: http://www-lium.univ-lemans.fr/sidekit/
#
# S4D is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
#
# S4D is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with SIDEKIT.  If not, see <http://www.gnu.org/licenses/>.

__author__ = 'meignier'

import sys
import os
from sidekit.features_extractor import FeaturesExtractor
from sidekit.features_server import FeaturesServer
import logging
import re
import numpy

def str2str_normalize(name):
    """
    removes accents and replace '_' by '_' the the string speaker
    :param name: the string to nomalize
    :return:
    """
    name = name.translate(str.maketrans('ÀÁÂÃÄÅàáâãäåÒÓÔÕÖØòóôõöøÈÉÊËèéêëÇçÌÍÎÏìíîïÙÚÛÜùúûüÿÑñ','AAAAAAaaaaaaOOOOOOooooooEEEEeeeeCcIIIIiiiiUUUUuuuuyNn')).lower()
    name = name.translate(str.maketrans("'",'_'))

    name = name.translate(str.maketrans('-','_'))
    return re.sub('_+','_',name)


def path_show_ext(fullpath, shortext=False):
    """
    splits a full file path into path, basename and extension
    :param fullpath: str
    :return: the path, the basename and the extension
    """
    tmp = os.path.splitext(fullpath)
    ext = tmp[1]
    p = tmp[0]
    if shortext == False:
        while tmp[1] != '':
            tmp = os.path.splitext(p)
            ext = tmp[1] + ext
            p = tmp[0]

    path = os.path.dirname(p)
    if path == '':
        path = '.'
    base = os.path.basename(p)
    return path, base, ext

def levenshtein_distance(s1, s2):
    if len(s1) > len(s2):
        s1, s2 = s2, s1
    distances = range(len(s1) + 1)
    for index2, char2 in enumerate(s2):
        new_distances = [index2 + 1]
        for index1, char1 in enumerate(s1):
            if char1 == char2:
                new_distances.append(distances[index1])
            else:
                new_distances.append(1 + min((distances[index1],
                                             distances[index1 + 1],
                                             new_distances[-1])))
        distances = new_distances
    return distances[-1]


def hms(s):
    """
    conversion of seconds into hours, minutes and secondes
    :param s:
    :return: int, int, float
    """
    h = int(s) // 3600
    s %= 3600
    m = int(s) // 60
    s %= 60
    return '{:d}:{:d}:{:.2f}'.format(h, m, s)


def get_feature_extractor(audio_filename_structure, type_feature_extractor):
    if type_feature_extractor == 'sid':
        fe = FeaturesExtractor(audio_filename_structure=audio_filename_structure,
             feature_filename_structure=None,
             sampling_frequency=16000,
             lower_frequency=133.3333,
             higher_frequency=6855.4976,
             filter_bank="log",
             filter_bank_size=40,
             window_size=0.025,
             shift=0.01,
             ceps_number=13,
             pre_emphasis=0.97,
             keep_all_features=True,
             vad='percentil',
             #vad=None,
             save_param=["energy", "cep", "vad"]
            )
    elif type_feature_extractor == 'sid8k':
        fe = FeaturesExtractor(audio_filename_structure=audio_filename_structure,
             feature_filename_structure=None,
             sampling_frequency=8000,
             lower_frequency=0,
             higher_frequency=4000,
             filter_bank="log",
             filter_bank_size=24,
             window_size=0.025,
             shift=0.01,
             ceps_number=12,
             pre_emphasis=0.95,
             keep_all_features=True,
             #vad='percentil',
             vad=None,
             save_param=["energy", "cep", "vad"]
            )
    elif type_feature_extractor == '8k' or type_feature_extractor == '8kcms'\
            or type_feature_extractor == '8ksns':
        fe = FeaturesExtractor(audio_filename_structure=audio_filename_structure,
             feature_filename_structure=None,
             sampling_frequency=8000,
             lower_frequency=0,
             higher_frequency=4000,
             filter_bank="log",
             filter_bank_size=24,
             window_size=0.025,
             shift=0.01,
             ceps_number=13,
             pre_emphasis=0.97,
             keep_all_features=True,
             #vad='percentil',
             vad=None,
             save_param=["energy", "cep", "vad"]
            )
    elif type_feature_extractor == 'basic':
        fe = FeaturesExtractor(audio_filename_structure=audio_filename_structure,
             feature_filename_structure=None,
             sampling_frequency=16000,
             lower_frequency=133.3333,
             higher_frequency=6855.4976,
             filter_bank="log",
             filter_bank_size=40,
             window_size=0.025,
             shift=0.01,
             ceps_number=13,
             pre_emphasis=0.97,
             keep_all_features=True,
             vad=None,
             save_param=["energy", "cep", "vad"]
            )
    else:
        logging.error('in get_feature_server, type_fe not found: ' + type_feature_extractor)
        return None
    return fe


def get_feature_server(filename_structure, feature_server_type):
    path, show, ext = path_show_ext(filename_structure)
    feature_filename_structure = None
    logging.info(path+' ## '+show+' ## '+ext)
    if ext.endswith('.h5') or ext.endswith('.hdf5'):
        feature_extractor = None
        feature_filename_structure = filename_structure
        logging.info('feature extractor --> None')
    else:
        audio_filename_structure = filename_structure
        feature_extractor = get_feature_extractor(audio_filename_structure, type_feature_extractor=feature_server_type)
    logging.info('-'*20)
    logging.info(feature_extractor)
    logging.info('-'*20)
    if feature_server_type == 'basic':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('energy', 'cep'),
                 keep_all_features=True)
    elif feature_server_type == 'sns':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('cep'),
                 delta=True,
                 keep_all_features=True)
    elif feature_server_type == 'sns_dnn':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('cep'),
                 delta=True,
                 context=(31, 31),
                 keep_all_features=True)
    elif feature_server_type == 'sid':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('energy', 'cep'),
                 feat_norm='cmvn_sliding',
                 delta=True,
                 double_delta=True,
                 keep_all_features=True)
    elif feature_server_type == 'sid8k':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('cep'),
                 feat_norm='cmvn_sliding',
                 delta=True,
                 double_delta=False,
                 keep_all_features=True)
    elif feature_server_type == '8k':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('cep'),
                 #delta=True,
                 keep_all_features=True)
    elif feature_server_type == '8ksns':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('cep'),
                 delta=True,
                 keep_all_features=True)
    elif feature_server_type == '8kcms':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('cep'),
                 feat_norm='cms',
                 #delta=True,
                 keep_all_features=True)
    elif feature_server_type == 'vad':
        feature_server = FeaturesServer(features_extractor=feature_extractor,
                 feature_filename_structure=feature_filename_structure,
                 dataset_list=('energy'),
                 keep_all_features=True)
    else:
        logging.error('in get_feature_server, feature_server_type not found: ' + feature_server_type)
        return None
    logging.info(feature_server)
    return feature_server

# def get_feature_server(input_dir='./{s}.h5', feature_server_type):
#     logging.info('get_feature_server type: '+feature_server_type)
#     if feature_server_type == 'diarization':
#         return FeaturesServer_test(input_dir=input_dir,
#                     config='diar_16k')
#     elif feature_server_type == 'sid':
#         return FeaturesServer_test(input_dir=input_dir,
#                     config='diar_16k', log_e=True, delta=True,
#                     double_delta=True, feat_norm='cms_sliding')
#     elif feature_server_type == 'sad':
#         return FeaturesServer_test(input_dir=input_dir,
#                     config='diar_16k', log_e=False, delta=True, double_delta=False)
#     else:
#         logging.error('in get_feature_server, feature_server_type not found: ' + feature_server_type)
#         return None


# def save_mfcc(diarization, audio_dir, mfcc_fn, feature_server_type):
#     fh = h5py.File(mfcc_fn, "w")
#     diar_out = diarization.copy_structure()
#     shows = diarization.make_index(['show'])
#
#     for show in shows:
#         # logging.info('mfcc: '+ show)
#         show_diar = shows[show]
#         model_iv = ModelIV()
#         feature_server = get_feature_server(audio_dir, feature_server_type=feature_server_type)
#         model_iv.set_feature_server(feature_server)
#         model_iv.set_diar(show_diar)
#         if feature_server_type == 'sid':
#             model_iv.vad()
#         else:
#             model_iv.diar_vad = show_diar
#
#         cep_full, _ = feature_server.load(show)
#         cluster_list = model_iv.diar_vad.make_index(['cluster'])
#         index = model_iv.diar_vad.features_by_cluster(show=show, cep_len=cep_full.shape[0])
#         for cluster in cluster_list:
#             logging.info('mfcc: '+show+' '+cluster)
#             mfcc_fn = show+'/'+cluster
#             cep = cep_full[index[cluster], :]
#             vad = numpy.ones(cep.shape[0])
#             diar_out.append(show=mfcc_fn, start=0, stop=cep.shape[0], cluster=cluster)
#             logging.info(cep.shape)
#             write_hdf5(mfcc_fn, fh, cep, None, None, None, label=vad)
#     return diar_out

# def save_mfcc(diarization, audio_dir='./', mfcc_fn='./out.h5', feature_server_type='sid'):
#     fh = h5py.File(mfcc_fn, "w")
#     shows = diarization.unique('show')
#     diar_out = diarization.copy_structure()
#     shows = diarization.make_index(['show'])
#     for show in shows:
#         # logging.info('mfcc: '+ show)
#         show_diar = shows[show]
#         model_iv = ModelIV()
#         feature_server = get_feature_server(audio_dir, feature_server_type=feature_server_type)
#         model_iv.set_feature_server(feature_server)
#         model_iv.set_diar(show_diar)
#         if feature_server_type == 'sid':
#             model_iv.vad()
#         else:
#             model_iv.diar_vad = show_diar
#         feature_server.load(show)
#         cep_full = feature_server.cep[0]
#         cluster_list = model_iv.diar_vad.make_index(['cluster'])
#         index = model_iv.diar_vad.features_by_cluster(show=show, cep_len=cep_full.shape[0])
#         for cluster in cluster_list:
#             logging.info('mfcc: '+show+' '+cluster)
#             mfcc_fn = show+'/'+cluster
#             cep = cep_full[index[cluster], :]
#             vad = numpy.ones(cep.shape[0])
#             diar_out.append(show=mfcc_fn, start=0, stop=cep.shape[0], cluster=cluster)
#             write_hdf5(mfcc_fn, fh, cep, label=vad)
#     return diar_out

class FeatureServerFake(FeaturesServer):
    def __init__(self, cep):
        self.cep = cep

    def load(self, show, channel=0, input_feature_filename=None, label=None, start=None, stop=None):
        return self.cep, numpy.ones(self.cep.shape[0], dtype='bool')

class FeatureServerCache(FeaturesServer):
    def __init__(self, featuresServer):
        self.shows = dict()
        self.featuresServer = featuresServer


    def load(self, show, channel=0, input_feature_filename=None, label=None, start=None, stop=None):
        key = show
        if label is not None:
            key += '##'+label
        #if start is not None:
        #    key += '##'+str(start)+'##'+str(stop)
        #for k in self.shows:
        #    logging.info('key: %s', k)
        if key in self.shows:
            #logging.info('load from mem '+key)
            cep = self.shows[key][start:stop,:]
            return cep, numpy.ones(cep.shape[0], dtype='bool')
        else:
            #logging.info('load from disque %s', key)
            cep, lbl = self.featuresServer.load(show, label=label, start=start, stop=stop)
            self.shows[key] = cep
            #logging.info('add: %s %d %d', key, (key in self.shows), True)
            return cep, lbl



import logging
import copy
from s4d.diar import Diar, Segment
from s4d.scoring import DER


def add_match(ref, hyp, assign, not_assigned):
    inv_map = {v: k for k, v in assign.items()}
    i = 0

    l = 0
    for cluster in ref.unique('cluster'):
        if len(cluster) > l:
            l = len(cluster)
    for cluster in hyp.unique('cluster'):
        if len(cluster) > l:
            l = len(cluster)
    f = 'Match {:03d} {:>'+str(l)+'s}'
    for key in sorted(inv_map, reverse=True):
        v = inv_map[key]
        r = v[0]
        h = v[1]
        nr = f.format(i, r)
        nh = f.format(i, h)
        ref.rename('cluster', [r], nr)
        hyp.rename('cluster', [h], nh)
        i += 1
    f = 'NotMatch {:>'+str(l)+'s}'
    for h in not_assigned:
        nh = f.format(h)
        hyp.rename('cluster', [h], nh)


def add_intersection(ref, hyp, assign):
        diar = Diar()
        if not diar._attributes.exist('color'):
            diar.add_attribut('color', (0.0, 0.0, 0.0, 1.0))
        map_ref = ref.make_index(['cluster'])
        map_hyp = hyp.make_index(['cluster'])
        for (lr, lh) in assign:
            for seg1 in map_ref[lr]:
                f = False
                stop = seg1['stop']
                for Seg2 in map_hyp[lh]:
                    if stop < Seg2['start']:
                        continue
                    inter1 = Segment.intersection(seg1, Seg2)
                    if inter1 is not None :
                        f = True
                        inter1['color'] = (0.6, 0.0, 1.0, 1.0)
                        inter1['cluster'] = seg1['cluster']
                        diar.append_seg(inter1)
                        inter2 = copy.deepcopy(inter1)
                        inter2['color'] = (0.0, 0.6, 1.0, 1.0)
                        inter2['cluster'] = Seg2['cluster']
                        diar.append_seg(inter2)
                    elif f:
                        break
        return diar


def add_prefix(diar, prefix):
    for seg in diar:
        seg['cluster'] = prefix+seg['cluster']


def add_collar(diar, collar):
    local_diar = Diar()
    local_diar.add_attribut('color', (0.2, 0.2, 0.2, 1.0))
    if local_diar._attributes.exist('type'):
        local_diar.add_attribut('type', 'collar')
    l = len(diar)
    for seg in diar:
        if l % 100 == 0:
            print(l)
        l -= 1
        show = seg['show']
        name = seg['cluster']
        start = seg['start']
        stop = seg['stop']
        local_diar.append(show=show, cluster=name, start=start - collar, stop=start+collar)
        local_diar.append(show=show, cluster=name, start=stop - collar, stop=stop+collar)
    diar.append_diar(local_diar)


def diar_diff(hyp, ref, match=True, inter=False, collar=25):
    hyp = copy.deepcopy(hyp)
    if not hyp._attributes.exist('color'):
        hyp.add_attribut('color', (0.0, 0.6, 0.0, 1.0))
    hyp.pack()

    ref = copy.deepcopy(ref)
    if not ref._attributes.exist('color'):
        ref.add_attribut('color', (0.6, 0.0, 0.0, 1.0))
    ref.pack()

    der = DER(hyp, ref)
    der.confusion()
    assign, not_assign = der.assignment()
    if match:
        add_match(ref, hyp, assign, not_assign)
    else:
        add_prefix(ref, 'REF: ')
        add_prefix(hyp, 'HYP: ')
        
    diar = copy.deepcopy(ref)
    diar.append_diar(hyp)
    
    if inter:
        logging.info('append intersection')
        diar.append_diar(add_intersection(ref, hyp, assign))

    if collar > 0:
        logging.info('append collar')
        add_collar(diar, collar)
    return diar


#!/usr/bin/env python
# -*- coding: utf-8 -*-
 
from setuptools import setup, find_packages
import s4d

setup(
    name='s4d',
    version=s4d.__version__,
    packages=find_packages(),
    author="Sylvain MEIGNIER",
    author_email="s4d@univ-lemans.fr",
    description="S4D: SIDEKIT for Diarization",
    long_description=open('README.md').read(),
    include_package_data=True,
    url='https://projets-lium.univ-lemans.fr/s4d/',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering",
    ],
)



__author__ = 'meignier'

import logging

from s4d.diar import Diar
import numpy as np
import pandas as pd
from s4d.clustering.hac_bic import GaussFull
from s4d.clustering.hac_utils import bic_square_root
import copy
import scipy

def sanity_check(cep, show, cluster='init'):
    """
    Removes equal MFCC of *cep* and return a diarization.

    :param cep: numpy.ndarry containing MFCC
    :param show: speaker of the show
    :return: a dirization object
    """
    table = Diar()

    # 1- diff on cep[i] - cep[i+1]
    # 2- sum of the n MFCC
    # 3- take equal values, give a boolean array
    b = np.sum(np.diff(cep, axis=0), axis=1) == 0
    # make a xor on the boolean array, true index+1 correspond to a boundary
    bits = b[:-1] ^ b[1:]
    # convert true value into a list of feature indexes
    # append 0 at the beginning of the list, append the last index to the list
    idx = [0] + (np.arange(len(bits))[bits] + 1).tolist() + [cep.shape[0]]
    # for each pair of indexes (idx[i] and idx[i+1]), create a segment
    for i in range(0, len(idx) - 1, 2):
        table.append(show=show, start=idx[i], stop=idx[i + 1], cluster=cluster)

    return table


def init_seg(cep, show='empty', cluster='init'):
    """
    Return an initial segmentation composed of one segment from the first to the
    last feature in *cep*.

    :param cep: numpy.ndarry containing MFCC
    :param show: the speaker of the cep
    :param cluster: str
    :return: a Diar object
    """
    length = cep.shape[0]
    table_out = Diar()
    table_out.append(show=show, start=0, stop=length, cluster=cluster)
    return table_out


def adjust(cep, diarization):
    """
    Moves the border of segment of *diarization* into lowest energy region and split
    segments greater than 30s

    :todo: change numpy.convolve to the panada version

    :param cep: a numpy.ndarray containing MFCC
    :param diarization: a Diarization object
    :return: a Diar object
    """
    energy_index = 0
    box = np.ones(100) / 100

    smooth = np.convolve(cep[:, energy_index], box, mode='same')
    adj_table = _adjust(smooth, diarization)
    return _split_e(smooth, adj_table, 30*100)


def _adjust(smooth, diarization, window_size=25):
    """
    The segment boundaries of *diarization* are moved slightly: segment start and
    segment stop will be located in low energy regions.

    :param smooth: sliding means of the energy (numpy.ndarry)
    :param diarization: the diarization object to adjust
    :param window_size: the half size of the zone to find the minimum energy around a
    border
    :return: a Diar object
    """
    diarization_out = copy.deepcopy(diarization)
    diarization_out.sort(['start'])
    prev = diarization_out[0]
    for i in range(1, len(diarization_out)):
        cur = diarization_out[i]
        start = cur['start']
        p = np.argmin(smooth[start - window_size:start + window_size])
        l1 = p + start - window_size - prev['start']
        l2 = prev['stop'] - p + start - window_size
        if l1 > 500 and l2 > 500:
            prev['stop'] = p + start - window_size
            cur['start'] = p + start - window_size
        prev = cur
    return diarization_out


def _split_e(smooth, diarization, split_size):
    """
    Long segments of *diarization* are  cut recursively at their points of lowest
    energy in order to yield segments shorter than *split_size* seconds.

    :param smooth: sliding means of the energy (numpy.ndarray)
    :param diarization: a Diarization object
    :param split_size: maximum size of a segment
    :return: a Diar object
    """
    diarization_out = Diar()
    for segment in diarization:
        _split_seg(smooth, segment, 250, split_size, diarization_out.segments)
    return diarization_out


def _split_seg(smooth, segment, min_seg_size, split_size, lst):
    """
    *segment*, a long segment, is cut recursively at their points of lowest energy
    in order to yield segments shorter than *split_size* seconds. The new
    segments greater than *min_seg_size* are appended into *lst*

    :param smooth: sliding means of the energy (numpy.ndarry)
    :param segment: a segment
    :param min_seg_size: minimum size of a segment
    :param split_size: maximum size of a segment
    :param lst: the new segments are added to this list
    :return:
    """
    stop = segment['stop'] - min_seg_size
    start = segment['start'] + min_seg_size
    l = segment['stop'] - segment['start']
    if l > split_size:
        m = start + np.argmin(smooth[start:stop])
        row_left = copy.deepcopy(segment)
        row_left['stop'] = m
        row_right = copy.deepcopy(segment)
        row_right['start'] = m
        _split_seg(smooth, row_left, min_seg_size, split_size, lst)
        _split_seg(smooth, row_right, min_seg_size, split_size, lst)
    else:
        lst.append(copy.deepcopy(segment))



def div_gauss(cep, show='empty', win=250, shift=0):
    """
    Segmentation based on gaussian divergence.

    The segmentation detects the instantaneous change points corresponding to
    segment boundaries. The proposed algorithm is based on the detection of
    local maxima. It detects the change points through a gaussian divergence
    (see equation below), computed using Gaussians with diagonal covariance 
    matrices. The left and right gaussians are estimated over a five-second 
    window sliding along the whole signal (2.5 seconds for each gaussian, 
    given *win* =250 features).
    A change point, i.e. a segment boundary, is present in the middle of the
    window when the gaussian divergence score reaches a local maximum.


        :math:`GD(s_l,s_r)=(\\mu_r-\\mu_l)^t\\Sigma_l^{-1/2}\\Sigma_r^{-1/2}(\\mu_r-\\mu_l)`

    where :math:`s_l` is the left segment modeled by the mean :math:`\mu_l` and
    the diagonal covariance matrix :math:`\\Sigma_l`, :math:`s_r` is the right
    segment modeled by the mean :math:`\mu_r` and the diagonal covariance
    matrix :math:`\\Sigma_r`.

    :param cep: numpy array of frames
    :param show: speaker of the show
    :param win: windows size in number of frames
    :return: a diarization object (s4d annotation)
    """

    length = cep.shape[0]
    # start and stop of the rolling windows A
    start_a = win - 1  # end of NAN
    stop_a = length - win
    # start and stop of the rolling windows B
    start_b = win + win - 1  # end of nan + delay
    stop_b = length

    # put features in a Pandas DataFrame
    df = pd.DataFrame(cep)
    # compute rolling mean and std in the window of size win, get numpy array
    # mean and std have NAN at the beginning and the end of the output array
    #mean = pd.rolling_mean(df, win).values
    #std = pd.rolling_std(df, win).values
    r = df.rolling(window=win, center=False)
    mean = r.mean().values
    std = r.std().values

    # compute GD scores using 2 windows A and B
    dist = (np.square(mean[start_a:stop_a, :] - mean[start_b:stop_b, :]) / (
        std[start_a:stop_a, :] * std[start_b:stop_b, :])).sum(axis=1)

    # replace missing value to match cep size
    dist_pad = np.lib.pad(dist, (win - 1, win), 'constant',
                          constant_values=(dist[0], dist[-1]))

    # remove non-speech frame
    # find local maximal at + or - win size
    borders = scipy.signal.argrelmax(dist_pad, order=win)[0].tolist()
    # append the first and last
    borders = [0] + borders + [length]

    diarization_out = Diar()
    spk = 0
    for i in range(0, len(borders) - 1):
        diarization_out.append(show=show, start=shift+borders[i],
                         stop=shift+borders[i + 1], cluster='S' + str(spk))
        spk += 1
    return diarization_out


def segmentation(cep, diarization, win_size=250):
    diarization_out = Diar()
    for segment in diarization:
        l = segment.duration()
        # logging.info('start: ', seg['start'],'end: ', seg['stop'], 'len: ', l)
        if l > 2 * win_size:
            cep_seg = segment.seg_features(cep)
            tmp = div_gauss(cep_seg, show=segment['show'], win=win_size, shift=segment['start'])
            diarization_out.append_diar(tmp)
        else:
            diarization_out.append_seg(segment)

    i=0
    for segment in diarization_out:
        segment['cluster'] = 'S'+str(i)
        i += 1

    return diarization_out


def bic_linear(cep, diarization, alpha, sr=False):
    """
    This segmentation over the signal fuses consecutive segments of the same
    speaker from the start to the end of the record.  The measure employs
    the :math:`\Delta BIC` based on Bayesian Information Criterion , using full
    covariance Gaussians (see :class:`gauss.GaussFull`), as defined in equation below.

        :math:`\\Delta BIC_{i,j} = PBIC_{i+j} - PBIC_{i} - PBIC_{j} -  P`

        :math:`PBIC_{x}  = \\frac{n_x}{2} \\log|\\Sigma_x|`

        :math:`cst  = \\frac{1}{2} \\alpha \\left(d + \\frac{d(d+1)}{2}\\right)`

        :math:`P  = cst \\times log(n_i+n_j)`

    where :math:`|\\Sigma_i|`, :math:`|\\Sigma_j|` and :math:`|\\Sigma|` are the
    determinants of gaussians associated to the left and right segments
    :math:`i`, :math:`j`
    and :math:`i+j`. :math:`\\alpha` is a parameter to set up. The penalty
    factor :math:`P` depends on :math:`d`, the dimension of the cep, as
    well as on :math:`n_i` and :math:`n_j`, refering to the total length of
    left segment :math:`i` and right segment :math:`j` respectively.

    if *sr* is True, BIC distance is replaced by the square root bic
    (see :py:func:`clustering.hac_utils.bic_square_root`)

    :param cep: numpy.ndarray
    :param diarization: a Diarization object
    :param alpha: the threshold
    :param sr: boolean
    :return: a Diar object
    """
    # logger = logging.getLogger(__name__)

    diarization_out = copy.deepcopy(diarization)
    diarization_out.sort(['show', 'start'])
    dim = cep.shape[1]
    cst = GaussFull.cst_bic(dim, alpha)

    if len(diarization) <= 1:
        return diarization_out
    segment1 = diarization_out[0];
    features1 = segment1.seg_features(cep)
    model1 = GaussFull(segment1['cluster'], dim)
    model1.add(features1)
    model1.compute()
    i = 1

    while i < len(diarization_out):
        segment2 = diarization_out[i];
        if segment2['start'] > segment1['stop']+1:
            # logging.warning('there is a hole between segment')
            i += 1
            segment1 = segment2
            continue
        features2 = segment2.seg_features(cep)
        model2 = GaussFull(segment2['cluster'], dim)
        model2.add(features2)
        model2.compute()

        model12 = GaussFull.merge(model1, model2)
        p = cst * np.log(model1.count + model2.count)
        if sr:
            p = bic_square_root(model1.count, model2.count, alpha, dim)
        delta_bic = model12.partial_bic - model1.partial_bic - model2.partial_bic - p
        #print(i, v, p)
        if delta_bic < 0.0:
            logging.debug('linear remove %s %s: %i/%i %f', model1.name, model2.name, i,
                          len(diarization_out), delta_bic)
            segment1['stop'] = segment2['stop']
            model1 = model12
            del diarization_out[i]
        else:
            logging.debug('linear next %s %s: %i/%i %f', model1.name, model2.name, i,
                          len(diarization_out), delta_bic)
            segment1 = segment2
            model1 = model2
            i += 1
    return diarization_out




import matplotlib
#matplotlib.use('TKAgg')

import matplotlib.pyplot as plot
from matplotlib.patches import Rectangle
from s4d.gui.player import AudioPlayer


class PlotDiar:
    """
    A viewer of segmentation
    """
    def __init__(self, seg, wav=None, title='', gui=False, pick=False, vgrid=False, size=(18, 9)):
        self.rect_picked = None
        self.rect_color = (0.0, 0.6, 1.0, 1.0)  # '#0099FF'
        self.rect_selected_color = (0.75, 0.75, 0, 1.0)  # 'y'

        plot.rcParams['keymap.fullscreen'] = 'ctrl+f'
        plot.rcParams['keymap.home'] = ''
        plot.rcParams['keymap.back'] = ''
        plot.rcParams['keymap.forward'] = ''
        plot.rcParams['keymap.pan'] = ''
        plot.rcParams['keymap.zoom'] = 'ctrl+z'
        plot.rcParams['keymap.quit'] = 'ctrl+q'
        plot.rcParams['keymap.grid'] = ''
        plot.rcParams['keymap.yscale'] = ''
        plot.rcParams['keymap.xscale'] = ''
        plot.rcParams['keymap.all_axes'] = ''
        plot.rcParams['toolbar'] = 'None'
        plot.rcParams['keymap.save'] = 'ctrl+s'
        # plot.rcParams.update({'font.family': 'courrier'})

        self.pick = pick
        self.gui = gui
        self.vgrid = vgrid
        self.fig = plot.figure(figsize=size, facecolor='white', tight_layout=True)
        self.title = title

        self.ax = self.fig.add_subplot(1, 1, 1)
        cids = list()
        if self.gui:
            cids.append(
                self.fig.canvas.mpl_connect('key_press_event', self._on_keypress))
            cids.append(
                self.fig.canvas.mpl_connect('button_press_event', self._on_click))
            if pick:
                cids.append(self.fig.canvas.mpl_connect('pick_event', self._on_pick))
        self.height = 5
        self.maxx = 0
        self.maxy = 0
        self.end_play = 0
        self.seg = seg
        self.wav = wav
        self.audio = None
        if self.wav is not None and self.gui:
            self.audio = AudioPlayer(wav)
            self.timer = self.fig.canvas.new_timer(interval=10)
            self.timer.add_callback(self._update_timeline)
            self.timer.start()

        self.timeline = self.ax.plot([0, 0], [0, 0], color='r')[-1]
        self.map = self.seg.make_index(['show', 'cluster'])
        self.time_stamp = list()
        self.time_stamp_idx = 0

    def _draw_timeline(self, t):
        """
        Draw the timeline a position t
        :param t: in second, a float

        """
        min, max = self.ax.get_ylim()
        self.timeline.set_data([t, t], [min, max])
        self._draw_info(t)

    def _update_timeline(self):
        """
        Update the timeline given the position in the audio player

        """
        if self.audio is not None and self.audio.playing():
            t = self.audio.time()
            min, max = self.ax.get_xlim()
            if t > self.end_play and self.rect_picked is not None:
                self.audio.pause()
                self.end_play = self.maxx
            self._draw_timeline(t)
            if t > max:
                self._dec_right(min, max)
            if t < min:
                self._dec_left(min, max)
            self.fig.canvas.draw()

    def _draw_info(self, t):
        """
        Draw information on segment and timestamp
        :param t: a float
        :return:
        """
        ch = 'time:{:s} ({:.3f} sec {:d} frame)'.format(self._hms(t), t,
                                                        int(t * 100))
        ch2 = '\n\n\n'
        if self.rect_picked is not None:
            s = self.rect_picked.get_x()
            w = self.rect_picked.get_width()
            e = s + w
            ch2 = 'segment  start: {:20s} ({:8.2f} sec {:8d} frame)\n'.format(
                self._hms(s), s, int(s * 100))
            ch2 += 'segment   stop: {:20s} ({:8.2f} sec {:8d} frame)\n'.format(
                self._hms(e), e, int(e * 100))
            ch2 += 'segment lenght: {:20s} ({:8.2f} sec {:8d} frame)\n'.format(
                self._hms(w), w, int(w * 100))

        plot.xlabel(ch + '\n' + ch2)

    def draw(self):
        """
        Draw the segmentation

        """
        y = 0
        labels_pos = []
        labels = []
        for show in self.map:
            for cluster in sorted(self.map[show].keys()):
                labels.append(cluster)
                labels_pos.append(y + self.height // 2)
                for row in self.map[show][cluster]:
                    x = row['start'] / 100
                    self.time_stamp.append(x)
                    self.time_stamp.append(row['stop'] / 100)
                    w = row.duration() / 100
                    self.maxx = max(self.maxx, row['stop'] / 100)
                    c = self.rect_color
                    if row._attributes.exist('color'):
                        c = row['color']
                    rect = plot.Rectangle((x, y), w, self.height,
                                          color=c, picker=self.pick)
                    self.ax.add_patch(rect)
                y += self.height
        if self.gui:
            plot.xlim([0, min(600, self.maxx)])
        else:
            plot.xlim([0, self.maxx])

        plot.ylim([0, y])
        plot.yticks(labels_pos, labels)
        self.maxy = y
        self.end_play = self.maxx
        for show in self.map:
            for cluster in self.map[show]:
                self.ax.plot([0, self.maxx], [y, y], linestyle=':',
                             color='#AAAAAA')
                y -= self.height

        plot.title(self.title + ' (last frame: ' + str(self.maxx) + ')')
        if self.gui:
            self._draw_info(0)
        plot.tight_layout()
        self.time_stamp = list(set(self.time_stamp))
        self.time_stamp.sort()

        if self.vgrid:
            for x in  self.time_stamp:
                self.ax.plot([x, x], [0, self.maxy], linestyle=':',
                             color='#AAAAAA')


    def _dec_right(self, min, max):
        """
        Move right
        :param min: a float
        :param max: a float

        """
        dec = (max - min) // 10
        diff = max - min
        if max + dec <= self.maxx:
            print('** 1 ', min, max, dec)
            plot.xlim(min + dec, max + dec)
        else:
            print('** 2 ', min, max, dec, diff)
            plot.xlim(self.maxx - diff, self.maxx)

    def _dec_left(self, min, max):
        """
        Move left
        :param min: a float
        :param max: a float

        """
        dec = (max - min) // 10
        diff = max - min
        if min - dec >= 0:
            plot.xlim(min - dec, max - dec)
        else:
            plot.xlim(0, diff)

    def _on_keypress(self, event):
        """
        manage the keypress event
        :param event: a key event

        """
        hmin, hmax = self.ax.get_xlim()
        diff = hmax - hmin
        if event.key == 'ctrl++' or event.key == 'ctrl+=':
            plot.xlim(hmin * 1.5, hmax * 1.5)
        elif event.key == 'ctrl+-':
            plot.xlim(hmin / 1.5, hmax / 1.5)
        elif event.key == 'escape':
            plot.xlim(0, self.maxx)
            plot.ylim(0, self.maxy)
        elif event.key == 'right':
            self._dec_right(hmin, hmax)
        elif event.key == 'left':
            self._dec_left(hmin, hmax)
        elif event.key == 'ctrl+right':
            plot.xlim(self.maxx - diff, self.maxx)
        elif event.key == 'ctrl+left':
            plot.xlim(0, diff)
        elif event.key == 'alt+right':
            self.time_stamp_idx = min(len(self.time_stamp)-1, self.time_stamp_idx + 1)
            if self.audio is not None:
                self.audio.pause()
                self.audio.seek(self.time_stamp[self.time_stamp_idx])
            self._draw_timeline(self.time_stamp[self.time_stamp_idx])
        elif event.key == 'alt+left':
            self.time_stamp_idx = max(0, self.time_stamp_idx - 1)
            if self.audio is not None:
                self.audio.pause()
                self.audio.seek(self.time_stamp[self.time_stamp_idx])
            self._draw_timeline(self.time_stamp[self.time_stamp_idx])
        elif event.key is None and self.audio is not None:
            self.audio.play()
        elif event.key == ' ' and self.audio is not None:
            self.audio.pause()

        self.fig.canvas.draw()

    def _on_click(self, event):
        """
        manage the mouse event
        :param event: a mouse event

        """
        if event.xdata is not None and self.rect_picked is None:
            if self.audio is not None:
                self.audio.pause()
                self.audio.seek(event.xdata)
            self._draw_timeline(event.xdata)
            self.fig.canvas.draw()

    def _on_pick(self, event):
        """
        manage the selection of a segment
        :param event: a picked event

        """
        if isinstance(event.artist, Rectangle) and event.mouseevent.dblclick:
            print('on pick dbclick')
            rect = event.artist
            x, y = rect.get_xy()
            w = rect.get_width()
            c = rect.get_fc()
            if self.rect_picked is not None:
                if self._colors_are_equal(c, self.rect_selected_color):
                    rect.set_color(self.rect_color)
                    self.rect_picked = None
                    self.end_play = self.maxx
                else:
                    self.rect_picked.set_color(self.rect_color)
                    rect.set_color(self.rect_selected_color)
                    self.rect_picked = rect
                    if self.audio is not None:
                        self.audio.pause()
                        self.audio.seek(x)
                    self.time_stamp_idx = self.time_stamp.index(x)
                    self.end_play = x + w
                    self._draw_timeline(x)
            else:
                rect.set_color(self.rect_selected_color)
                self.rect_picked = rect

                if self.audio is not None:
                    self.audio.pause()
                    self.audio.seek(x)
                self.time_stamp_idx = self.time_stamp.index(x)
                self.end_play = x + w
                self._draw_timeline(x)

            self.fig.canvas.draw()

    @classmethod
    def _colors_are_equal(cls, c1, c2):
        """
        Compare two colors
        """
        for i in range(4):
            if c1[i] != c2[i]:
                return False
        return True

    @classmethod
    def _hms(cls, s):
        """
        conversion of seconds into hours, minutes and secondes
        :param s:
        :return: int, int, float
        """
        h = int(s) // 3600
        s %= 3600
        m = int(s) // 60
        s %= 60
        return '{:d}:{:d}:{:.2f}'.format(h, m, s)



import numpy as np
import logging
import copy
from sidekit import Mixture, FeaturesServer
from s4d.clustering.hac_utils import argmin, roll
from s4d.diar import Diar
from sidekit.statserver import StatServer
from bottleneck import argpartition

class HAC_CLR:
    """
    CLR Hierarchical Agglomerative Clustering (HAC) with GMM trained by MAP
    """
    def __init__(self, features_server, diar, ubm, ce=False, ntop=5):
        assert isinstance(features_server, FeaturesServer), 'First parameter has to be a FeatureServer'
        assert isinstance(diar, Diar), '2sd parameter has to be a Diar (segmentationContener)'
        assert isinstance(ubm, Mixture), '3rd parameter has to be a Mixture'

        self.features_server = features_server
        self.diar = copy.deepcopy(diar)
        self.merge = []
        self.nb_merge = 0
        self.ubm = ubm
        self.ce = ce
        self.stat_speaker = None
        self.stat_seg = None
        self.llr = None
        self.ntop = ntop
        #self.init_train()
        #self._init_distance()

    def _get_cep(self, map, cluster):
        cep_list = list()
        for show in map[cluster]:
            idx = self.diar.features_by_cluster(show)[cluster]
            if len(idx) > 0:
                tmp, vad = self.features_server.load(show)
                cep_list.append(tmp[0][idx])
        cep = np.concatenate(cep_list, axis=0)
        return cep

    def _ll(self, ubm, cep, mu=None, name='ubm', argtop = None):
        # ajouter le top gaussien
        lp = ubm.compute_log_posterior_probabilities(cep, mu=mu)

        if argtop is None:
            #logging.info('compute argtop '+speaker)
            argtop = argpartition(lp*-1.0 , self.ntop, axis=1)[:, :self.ntop]
            #logging.info(argtop.shape)
        if self.ntop is not None:
            #logging.info('use ntop '+speaker)
            #logging.info(argtop.shape)
            #logging.info(lp.shape)
            lp = lp[np.arange(argtop.shape[0])[:, np.newaxis], argtop]

        # ppMax = numpy.max(lp, axis=1)

        ll = np.log(np.sum(np.exp(lp), axis=1))
        # ll = ppMax + numpy.log(numpy.sum(numpy.exp((lp.transpose() - ppMax).transpose()),
        #                    axis=1))
        not_finite = np.logical_not(np.isfinite(ll))
        cpt = np.count_nonzero(not_finite)
        # ll[finite] = numpy.finfo('d').min
        ll[not_finite] = 1.0e-200
        m = np.mean(ll)
        if cpt > 0:
            logging.info('model ' + name + '), nb trame with llk problem: %d/%d \t %f', cpt, cep.shape[0], m)
        return m, argtop

    def initial_models(self, nb_threads=1):
        # sort by show to minimize the reading of mfcc by the statServer
        self.diar.sort(['show'])
        # Compute statistics by segments
        self.stat_seg = StatServer(self.diar.id_map())
        self.stat_seg.accumulate_stat(self.ubm, self.features_server)
        self.stat_speaker = self.stat_seg.adapt_mean_MAP_multisession(self.ubm)

    def initial_distances(self, nb_threads=1):
        map = self.diar.make_index(['cluster', 'show'])
        nb = self.stat_speaker.modelset.shape[0]

        self.llr = np.full((nb, nb), np.nan)
        self.dist = np.full((nb, nb), np.nan)
        for i, name_i in enumerate(self.stat_speaker.modelset):
            cep_i = self._get_cep(map, name_i)
            argtop = None
            ll_ubm = None
            if self.ntop is not None or self.ce == False:
                ll_ubm, argtop = self._ll(self.ubm, cep_i, argtop=argtop)

            # self.merge.append([])
            for j, name_j in enumerate(self.stat_speaker.modelset):
                mu = self.stat_speaker.get_model_stat1_by_index(j)
                # if i == 0:
                #    logging.debug(mu)
                self.llr[i, j], _ = self._ll(self.ubm, cep_i, mu=mu, name=name_j, argtop=argtop)
            if self.ce:
                self.llr[i,:] -= self.llr[i,i]
            else:
                self.llr[i,:] -= ll_ubm

        # logging.debug(self.llr)
        self.dist = (self.llr + self.llr.T)*-1.0
        np.fill_diagonal(self.dist, np.finfo('d').max)

    def update(self, i, j, nb_threads=1):
        name_i = self.stat_speaker.modelset[i]
        name_j = self.stat_speaker.modelset[j]
        # logging.debug('%d %d / %s %s', i, j, name_i, name_j)

        for k in range(len(self.stat_seg.modelset)):
            if self.stat_seg.modelset[k] == name_j:
                self.stat_seg.modelset[k] = name_i

        self.stat_speaker = self.stat_seg.adapt_mean_MAP_multisession(self.ubm)

        self.llr = roll(self.llr, j)

        self.diar.rename('cluster', [name_j], name_i)
        map = self.diar.make_index(['cluster', 'show'])
        cep_i = self._get_cep(map, name_i)
        argtop = None
        ll_ubm = None
        if self.ntop > 0 or self.ce == False:
            ll_ubm, argtop = self._ll(self.ubm, cep_i, argtop=argtop)
        for k, name_k in enumerate(self.stat_speaker.modelset):
            mu = self.stat_speaker.get_model_stat1_by_index(k)
            self.llr[i, k], _ = self._ll(self.ubm, cep_i, mu=mu, name=name_k)
        if self.ce:
            self.llr[i,:] -= self.llr[i,i]
        else:
            self.llr[i,:] -= ll_ubm

        self.dist = (self.llr + self.llr.T)*-1.0
        np.fill_diagonal(self.dist, np.finfo('d').max)

    def information(self, i, j, value):
        models = self.stat_speaker.modelset
        self.merge.append([self.nb_merge, models[i], models[j], value])

    def perform(self, thr = 0.0, to_the_end=False):
        models = self.stat_speaker.modelset
        nb = len(models)
        self.nb_merge = -1
        for i in range(nb):
            self.information(i, i, 0)

        i, j, v = argmin(self.dist, nb)
        self.nb_merge = 0
        while v < thr and nb > 1:
            self.information(i ,j, v)
            self.nb_merge += 1
            logging.debug('merge: %d c1: %s (%d) c2: %s (%d) dist: %f',
                          self.nb_merge, models[i], i, models[j], j, v)
            # update merge
            # update model and distance
            self.update(i, j)
            nb -= 1
            i, j, v = argmin(self.dist, nb)

        end_diar = copy.deepcopy(self.diar)
        if to_the_end:
            while nb > 1:
                self.information(i ,j, v)
                self.nb_merge += 1
                logging.debug('merge: %d c1: %s (%d) c2: %s (%d) dist: %f',
                              self.nb_merge, models[i], i, models[j], j, v)
                # update merge
                # update model and distance
                self.update(i, j)
                nb -= 1
                i, j, v = argmin(self.dist, nb)

        return end_diar



